;; Generated package description from shell-command+.el  -*- no-byte-compile: t -*-
(define-package "shell-command+" "2.0.0.0.20200927.160257" "An extended shell-command" '((emacs "24.1")) :keywords '("unix" "processes" "convenience") :authors '(("Philip K." . "philipk@posteo.net")) :maintainer '("Philip K." . "philipk@posteo.net") :url "http://elpa.gnu.org/packages/shell-command+.html")
